use std::{
    any::{Any, TypeId},
    collections::HashMap,
};

use utilities::prelude::{remove_life_time, remove_life_time_mut};

type Untyped = dyn Any + Send + Sync;

#[derive(Default)]
pub struct Resources {
    map: HashMap<TypeId, Box<Untyped>>,
}

impl Resources {
    pub fn insert<T: Any + Send + Sync>(&mut self, value: T) -> Option<T> {
        self.map
            .insert(TypeId::of::<T>(), Box::new(value))
            .map(|any| *Self::downcast_unchecked(any))
    }

    pub fn insert_if_not_exists<T: Default + Any + Send + Sync>(&mut self) {
        if !self.contains::<T>() {
            self.insert(T::default());
        }
    }

    pub fn remove<T: Any + Send + Sync>(&mut self) -> Option<T> {
        self.map
            .remove(&TypeId::of::<T>())
            .map(|any| *Self::downcast_unchecked(any))
    }

    pub fn get<T: Any + Send + Sync>(&self) -> &T {
        self.get_opt::<T>().unwrap()
    }

    pub fn get_by_type_id<T: Any + Send + Sync>(&self, type_id: TypeId) -> &T {
        self.get_opt_by_type_id(type_id).unwrap()
    }

    pub fn get_unchecked<'a, T: Any + Send + Sync>(&self) -> &'a T {
        unsafe { remove_life_time(self.get::<T>()) }
    }

    pub fn get_opt<T: Any + Send + Sync>(&self) -> Option<&T> {
        self.get_opt_by_type_id(TypeId::of::<T>())
    }

    pub fn get_opt_by_type_id<T: Any + Send + Sync>(&self, type_id: TypeId) -> Option<&T> {
        debug_assert_eq!(type_id, TypeId::of::<T>());

        self.map
            .get(&type_id)
            .map(|any| Self::downcast_ref_unchecked(any))
    }

    pub fn get_mut<T: Any + Send + Sync>(&mut self) -> &mut T {
        self.get_mut_opt::<T>().unwrap()
    }

    pub fn get_mut_by_type_id<T: Any + Send + Sync>(&mut self, type_id: TypeId) -> &mut T {
        self.get_mut_opt_by_type_id(type_id).unwrap()
    }

    pub fn get_mut_by_type_id_untyped(&mut self, type_id: TypeId) -> &mut Untyped {
        self.get_mut_opt_by_type_id_untyped(type_id).unwrap()
    }

    pub fn get_mut_unchecked<'a, T: Any + Send + Sync>(&mut self) -> &'a mut T {
        unsafe { remove_life_time_mut(self.get_mut::<T>()) }
    }

    pub fn get_mut_opt<T: Any + Send + Sync>(&mut self) -> Option<&mut T> {
        self.get_mut_opt_by_type_id(TypeId::of::<T>())
    }

    pub fn get_mut_opt_by_type_id<T: Any + Send + Sync>(
        &mut self,
        type_id: TypeId,
    ) -> Option<&mut T> {
        debug_assert_eq!(type_id, TypeId::of::<T>());

        self.map
            .get_mut(&type_id)
            .map(|any| Self::downcast_mut_unchecked(any))
    }

    pub fn get_mut_opt_by_type_id_untyped(&mut self, type_id: TypeId) -> Option<&mut Untyped> {
        self.map.get_mut(&type_id).map(|any| any.as_mut())
    }

    pub fn contains<T: Any + Send + Sync>(&self) -> bool {
        self.map.contains_key(&TypeId::of::<T>())
    }
}
